/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.util.TreeSet;
import java.util.HashSet;

/**
 * From: http://www.gamasutra.com/features/19990212/sm_01.htm
 */
public class AStarSearch
{
	static class Node
	implements Comparable
	{
		final int x;
		final int y;
		//private int cost; // "f" // XXX Need not be stored, is sum of next two.
		private int costToHere; // "g"
		private int costToGoal; // "h"
		Node parent; // for backtracking

		Node(int x, int y)
		{
			this.x = x;
			this.y = y;
			//this.cost = -1;
			this.costToHere = -1;
			this.costToGoal = -1;
			this.parent = null;
		}

		public int cost()
		{
			return this.costToHere + this.costToGoal;
		}

		public String toString()
		{
			if (this.cost() == -1)
				return "(" +x+ "," +y+ ")";
			return "(" +x+ "," +y+ ")@" +hashCode() 
				+ " [" +this.cost()+ ", here=" +this.costToHere+ ", goal=" +this.costToGoal+ "]";
		}

		public boolean equals(Object o)
		{
			return this == o;
		}

		public int hashCode()
		{
			return System.identityHashCode(this);
		}

		// Need to return <0 if this is cheaper than other
		public int compareTo(Object o)
		{
			Node other = (Node)o;

			// Only same if identical
			if ((other.x == this.x)
			    && (other.y == this.y))
				return 0;
			
			if (this.cost() == other.cost())
			{
				// If same cost, then arbitrary sorting by dist from origin...
				if (this.x == other.x)
					return this.y - other.y;
				return this.x - other.x;
			}
			
			return this.cost() - other.cost();
		}
	}

	private final Node[] nodes; // all nodes, keeps `em unique
	private final TreeSet open;  // nodes to be considered
	private final HashSet closed; // nodes with all neighbors considered
	private final Node start;
	private final Node goal;
	private final Board board;

	private Node getNode(int x, int y)
	{
		int width = this.board.width();
		int offset = (y-1)*width + (x-1);
		Node n = nodes[offset];
		if (n == null)
		{
			n = new Node(x,y);
			nodes[offset] = n;
		}
		else
		{
			if ((n.x != x) || (n.y != y))
				throw new Error("Bogus Node in my nodes list.");
		}
		return n;
	}

	private int goalCostEstimate(Node n)
	{
		int dx = n.x - this.goal.x;
		int dy = n.y - this.goal.y;
		if (dx < 0)
			dx = -dx;
		if (dy < 0)
			dy = -dy;
		return dx+dy;
	}

	public AStarSearch(Board board,
			   int startX, int startY,
			   int destX, int destY)
	{
		this.board = board;
		this.open = new TreeSet(); // of Nodes
		this.closed = new HashSet(); // of Nodes
		this.nodes = new Node[board.width() * board.height()];

		this.goal = getNode(destX, destY);
		this.start = getNode(startX, startY);

		Trace.t(Trace.ASTAR, "Looking for path from " +start+ " to " +goal);

		this.start.costToHere = 0;
		this.start.costToGoal = goalCostEstimate(start);
		//this.start.cost = this.start.costToGoal + this.start.costToHere;
		this.start.parent = null;
		open.add(this.start);
	}
	
	private void checkNeighbor(Node current, Direction d)
	{
		if (! this.board.impassableIn(current.x, current.y, d))
		{
			Node n = getNode(current.x + d.dx(), current.y + d.dy());

			// XXX add swamp aversion (+ this.board.travelPenalty(n.x, n.y)
			int newCostToHere = current.costToHere + 1;
			
			// Get existing open/closed state for this sort of node
			boolean inOpen = open.contains(n);
			boolean inClosed = closed.contains(n);
			
			if ((inOpen || inClosed)
			    && (n.costToHere <= newCostToHere))
				return; // New variation is not useful
			
			if (inOpen)
				open.remove(n);  // Must remove as sorting position may have changed

			// New variation is better, update node info:
			n.parent = current;
			n.costToHere = newCostToHere;
			n.costToGoal = goalCostEstimate(n); // XXX only if (n.costToGoal == -1)
			//n.cost = n.costToHere + n.costToGoal;
			if (inClosed)
				closed.remove(n);
			open.add(n);
		}
		else
		{
			Trace.t(Trace.ASTAR, "Neighbor " +d+ " of " +current+ " impassable.");
		}
	}
			
	private Direction handleSuccess(Node n)
	{
		int pathlen = 0;
		Node prev = null;
		Trace.t(Trace.ASTAR, "Found path:");
		while (n.parent != null)
		{
			Trace.t(Trace.ASTAR, "  " +n);
			pathlen++;
			prev = n;
			n = n.parent;
		}
		Trace.t(Trace.ASTAR, "  (start = " +this.start+ ")");
		Trace.t(Trace.ASTAR, "prev=" +prev);
		Trace.t(Trace.ASTAR, "Path length = " +pathlen);
		
		if (n != this.start)
			throw new Error("Hmm... traced back to " +n+ " not start (" +this.start+ ")");
		
		if ((prev == null)
		    && (this.start != this.goal))
			throw new Error("No prev, yet there is a path.");
		
		int dx = this.start.x - prev.x;
		if ((dx != 1)
		    && (dx != 0)
		    && (dx != -1))
			throw new Error("Should be single step from " +this.start+ " to " +prev);
		
		int dy = this.start.y - prev.y;
		if ((dy != 1)
		    && (dy != 0)
		    && (dy != -1))
			throw new Error("Should be single step from " +this.start+ " to " +prev);
		
		return Direction.fromTo(this.start.x, this.start.y,
					prev.x, prev.y);
	}

	public Direction search()
	{
		int max = 1000;
		
		while(! open.isEmpty())
		{
			Node bestGuess = (Node)(this.open.first());
			this.open.remove(bestGuess);
			// Tried all of bestGuess
			this.closed.add(bestGuess);

			Trace.t(Trace.ASTAR, "Considering " +bestGuess);

			if (bestGuess == this.goal)
				return handleSuccess(bestGuess);

			checkNeighbor(bestGuess, Direction.North);
			checkNeighbor(bestGuess, Direction.South);
			checkNeighbor(bestGuess, Direction.East);
			checkNeighbor(bestGuess, Direction.West);

		}

		// No such path!
		return null;
	}
}
