/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeSet;
import java.lang.Comparable;

// Collection of packages, can lookup by specific dest location, or by
// dest location with most value
public final class PackageSet
{
	private static final class RowHash extends IntKeyedHashtable
	implements IntKeyedHashtable.IntKey
	{
		private final int row;

		RowHash(int row)
		{
			this.row = row;
		}

		public int keyInt()
		{
			return this.row;
		}
	}

	public static final class PackageList extends HashSet
	implements IntKeyedHashtable.IntKey, Comparable
	{
		private final int x;
		private int value;

		PackageList(int x)
		{
			super(10);
			this.x = x;
			this.value = 0;
		}

		public int value()
		{
			return this.value;
		}

		public boolean add(Package p)
		{
			if (p.weight() == -1)
				throw new Error("Don't add packages you don't know the weight for!");

			if (p.destX() != this.x)
				throw new Error("Inserting " +p+ " in wrong " +this);
			boolean unique = super.add(p);
			if (unique)
				this.value += p.weight();
			return unique;
		}

		public boolean remove(Package p)
		{
			if (p.destX() != this.x)
				throw new Error("Removing " +p+ " in wrong " +this);
			boolean done = super.remove(p);
			if (done)
				this.value -= p.weight();
			return done;
		}
		
		public int compareTo(Object o)
		{
			if (o == this)
				return 0;

			PackageList other = (PackageList)o;

			if (other.value == this.value)
			{
				// Only return 0 if identical objects.

				if (other.hashCode() < this.hashCode())
					return -1;
				return 1;
			}

			// If other is worth more, return >0
			return other.value - this.value;
		}

		public boolean equals(Object o)
		{
			return this == o;
		}
		
		public int hashCode()
		{
			return System.identityHashCode(this);
		}

//		public String toString()
//		{
//			return "[PackageList@" +hashCode()+ ":" +super.toString() +"]";
//		}

		public int keyInt()
		{
			return this.x;
		}
	}


	private final IntKeyedHashtable rows = new IntKeyedHashtable(50);

	private final TreeSet usefulLists = new TreeSet();
	int pkgCount = 0;

	public PackageSet()
	{
	}

	private RowHash rowHashFor(int y)
	{
		RowHash r = (RowHash)rows.get(y);
		if (r == null)
		{
			r = new RowHash(y);
			rows.add(r);
		}
		return r;
	}

	private PackageList packageListFor(int x, int y)
	{
		RowHash r = rowHashFor(y);
		PackageList pl = (PackageList)r.get(x);
		if (pl == null)
		{
			pl = new PackageList(x);
			r.add(pl);
		}
		return pl;
	}

	public void add(Package p)
	{
		int x = p.destX();
		int y = p.destY();

		Trace.t(Trace.PACKAGESET, "add("+p+")");
		
		PackageList pl = packageListFor(x,y);
		Trace.t(Trace.PACKAGESET, "   add("+p+") pl=" +pl);
		usefulLists.remove(pl); // Re-insert as pl value changes
		boolean unique = pl.add(p);
		if (unique)
		{
			Trace.t(Trace.PACKAGESET, "   add("+p+") is new pkg.");
			pkgCount++;
		}
		usefulLists.add(pl);
		//System.err.println("++UsefulLists= " +usefulLists);
	}

	// can return null
	private RowHash getRowHash(int y)
	{
		return (RowHash)rows.get(y);
	}

	// can return null
	public PackageList getPackageList(int x, int y)
	{
		RowHash r = getRowHash(y);
		if (r == null)
			return null;
		PackageList pl = (PackageList)r.get(x);
		return pl;
	}
	
	public boolean remove(Package p)
	{
		PackageList pl = getPackageList(p.destX(), p.destY());
		if (pl == null)
			return false;
		usefulLists.remove(pl); // re-insert as pl value changes
		boolean removed = pl.remove(p);
		if (removed)
		{
			pkgCount--;
		}
		usefulLists.add(pl);
		//System.err.println("--UsefulLists= " +usefulLists);
		return removed;
	}

	public boolean isEmpty()
	{
		return (this.pkgCount == 0);
	}

	public int packageCount()
	{
		return this.pkgCount;
	}

	public String toString() 
	{
		return "PackageSet[\n   rows=" +rows+ "\n"
			+ "   useful=" +usefulLists+ "\n]";
	}

	// Find PackageList from sparse array with highest value...
	public Package bestToDeliver()
	{
		int max = 12;
		if (usefulLists.isEmpty())
			return null;

		//System.err.println("bestToDeliver: this=" +this);

		PackageList bestList = (PackageList)(usefulLists.first());
		//System.err.println("bestToDeliver: bestList=" +bestList);
		return (Package)(bestList.iterator().next());
	}
}
