/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.net.*;
import java.io.*;
import java.util.ArrayList;

public final class Server
{
	private final String host;
	private final int port;
	private final SimpleToker input;
	private final PrintWriter output;

	public Server(String host, int port)
	{
		final int INPUT_BUF_SIZE=1024;

		this.host = host;
		this.port = port;

		try
		{
			Socket s = new Socket(host,port);
			this.input = new SimpleToker(new InputStreamReader(s.getInputStream()), 1024);
			this.output = new PrintWriter(new OutputStreamWriter(s.getOutputStream()), false);
		}
		catch (IOException ioe)
		{
			throw new Error("socket setup " +host+ ":" +port, ioe);
		}
	}

	public void establish()
	{
		sendMessage("Player");
	}

	void flush()
	{
		this.input.flush();
	}

	public void close()
	{
		try
		{
			this.input.close();
			this.output.close();
		}
		catch (IOException ioe)
		{
			System.err.println("ioe during close... :" +ioe);
		}
	}


	public void getBoard(Board b)
	{
		Trace.t(Trace.SERVER, "getBoard...");
		//String[] toks = getLine(2);
		//int width = Integer.parseInt(toks[0]);
		//int height =  Integer.parseInt(toks[1]);
		int width = this.input.readInt();
		int height = this.input.readInt();
		this.input.lineDone();
		// First line of text from server, thus important
		System.out.println("Board is " +width+ "x" +height);
		b.setSize(width, height);
		for (int i = 0; i < height; i++)
		{
			String s = getBoardLine(width);
			b.setRow(i, s);
		}
	}

	private void sendMessage(String s)
	{
		Trace.t(Trace.SEND_MSG, "Sending :" +s);
		this.output.println(s);
		this.output.flush();
	}

	public String getBoardLine(int width)
	{
		String s = this.input.readLine();
		this.input.lineDone();
		if (s.length() != width)
			throw new Error("Expecting board input.  Got: " +s);
		return s;
	}

	private static final Package[] noPackages = new Package[0];

	public Package[] getCurrentPackages()
	{
		Trace.t(Trace.SERVER, "getPackages...");
		ArrayList packages = new ArrayList(20);
		do {
			if (! this.input.testEndOfLine())
			{
				int id = this.input.readInt();
				int x  = this.input.readInt();
				int y  = this.input.readInt();
				int w  = this.input.readInt();
			
				packages.add(new Package(id, x, y, w));
			}
		} while (! this.input.endOfLine());
		this.input.lineDone();
		
		return (Package[])packages.toArray(noPackages);
	}

	public void sendCommand(Command c)
	{
		Trace.t(Trace.SEND_CMD, "Sending command " +c);
		try
		{
			c.writeTo(this.output);
			this.output.flush();
		}
		catch (IOException ioe)
		{
			throw new Error("Writing command " +c, ioe);
		}
	}

//	private static final String[] noStrings = new String[0];
//	public String[] getLine(int expectedTokens)
//	{
//		try
//		{
//			String l = this.input.readLine();
//
//			showLine(l);
//
//			if (l == null)
//				throw new EndOfGame("Got null from server.");
//			String[] s = l.split("\\s");
//			if (s == null)
//				throw new Error("split returned null??");
//			if ((expectedTokens != -1) 
//			    && (s.length != expectedTokens))
//				throw new Error("Unexpected token count.  Got " +s.length);
//			if ((s.length == 1) && s[0].equals(""))
//				return noStrings;
//			return s;
//		}
//		catch (IOException ioe)
//		{
//			throw new Error("Expecting a line of text.", ioe);
//		}
//	}
	
	public void getMyState(Robot me)
	{
		Trace.t(Trace.SERVER, "getMyState...");
//		String[] s = getLine(3);
//		me.setId(Integer.parseInt(s[0]));
//		me.setCap(Integer.parseInt(s[1]));
//		me.setMoney(Integer.parseInt(s[2]));
		me.setId(this.input.readInt());
		me.setCap(this.input.readInt());
		me.setMoney(this.input.readInt());
		this.input.lineDone();
	}

	public void updateWorldState(WorldState worldState)
	{
		Trace.t(Trace.SERVER, "updateWorldState...");
		int roboid = -1;
		do {
			if (this.input.testEndOfLine())
				throw new EndOfGame("??");

			String c = this.input.readToken();
			if ((c == null)
			    || (c.equals("")))
				throw new Error("expecting a useful token.");
			
			if (c.charAt(0) == '#')
			{
				roboid = Integer.parseInt(c.substring(1));
				Trace.t(Trace.UPDATEWORLD, "Robot=" +roboid);
			}
			else if (c.equals("P"))
			{
				int id = this.input.readInt();
				Trace.t(Trace.UPDATEWORLD, "  Picked up " +id);
				// picked up
				worldState.robotPicked(roboid, id);
			}
			else if (c.equals("D"))
			{
				int id = this.input.readInt();
				Trace.t(Trace.UPDATEWORLD, "  Dropped " +id);
				// dropped pkg
				worldState.robotDropped(roboid, id);
			}
			else if (c.equals("X"))
			{
				// set location
				int x = this.input.readInt();
				
				c = this.input.readToken();
				if (! c.equals("Y"))
					throw new Error("Expected 'Y' in coord update, got: " +c);
				
				int y = this.input.readInt();
				
				Trace.t(Trace.UPDATEWORLD, "  Placed at (" +x+ "," +y+ ")");
				worldState.robotPlaced(roboid, x, y);
			}
			else
			{
				Direction d = Direction.forstr(c);
				Trace.t(Trace.UPDATEWORLD, "  Moved " +d);
				
				// Throws error if not one of 'NSWE':
				worldState.robotMoved(roboid, d);
			}
		} while (! this.input.endOfLine());
		this.input.lineDone();
	}
}

