/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.io.*;

public final class SimpleToker
{
	private final PushbackReader input;
	private final StringBuffer sbuf;
	private boolean eol;

	SimpleToker(InputStreamReader isr, int bufsize)
		throws IOException
	{
		this.input = new PushbackReader(new BufferedReader(isr, bufsize));
		this.sbuf = new StringBuffer(3000);
		this.eol = false;
	}
	
	void close()
		throws IOException
	{
		this.input.close();
	}
	
	String readLine()
	{
		try
		{
			
			this.sbuf.setLength(0);
			while (true)
			{
				int ci = this.input.read();
				if (ci == -1)
				{
					Trace.t(Trace.TOKER, "readline(eof): `" +sbuf+ "'");
					return sbuf.toString();
				}
				char c = (char)ci;
				if (c == '\n')
				{
					this.eol = true;
					Trace.t(Trace.TOKER, "readline(\\n): `" +sbuf+ "'");
					return sbuf.toString();
				}
				this.eol = false;
				sbuf.append(c);
			}
		}
		catch (IOException ioe)
		{
			throw new Error("Gah.", ioe);
		}
	}
	
	int readInt()
	{
		try
		{
			if (this.eol)
				throw new Error("Expecting int, but at eol.");

			int val = 0;
			boolean hit = false;
			while (true)
			{
				int ci = this.input.read();
				Trace.t(Trace.TOKER, "  readInt(ci=" +ci+ "), val=" +val);
				if (ci == -1)
				{
					if (!hit)
						throw new Error("Unexpected EOF");
					return val;
				}
				char c = (char)ci;
				if (c == '\n')
				{
					this.eol = true;
					Trace.t(Trace.TOKER, "readInt(\\n): " +val);
					return val;
				}
				if (Character.isWhitespace(c))
				{
					this.eol = false;
					if (hit)
					{
						Trace.t(Trace.TOKER, "readInt(ws): " +val);
						return val;
					}
				}
				else if (Character.isDigit(c))
				{
					this.eol = false;
					hit = true;
					val = val*10 + (c - '0');
				}
				else // non-white, non-numeric
				{
					if (!hit)
					{
						this.input.unread(ci);
						char[] buf = new char[1000];
						this.input.read(buf, 0, buf.length);
						throw new Error("Unexpected non-digits:\n" +new String(buf));
					}
					this.eol = false;
					this.input.unread(ci);
					Trace.t(Trace.TOKER, "readInt(nws): " +val);
					return val;
				}
			}
		}
		catch (IOException ioe)
		{
			throw new Error("Gah.", ioe);
		}
		
	}
	
	String readToken()
	{
		try
		{
			if (this.eol)
				return null;

			sbuf.setLength(0);
			while (true)
			{
				int ci = this.input.read();
				Trace.t(Trace.TOKER, "  readToken(ci=" +ci+ "), buf=`" +sbuf+ "'");
				if (ci == -1)
				{
					Trace.t(Trace.TOKER, "readToken(eof): `" +sbuf+ "'");
					if (sbuf.length() == 0)
						throw new Error("Unexpected EOF.");
					return sbuf.toString();
				}
				char c = (char)ci;
				if (c == '\n')
				{
					this.eol = true;
					Trace.t(Trace.TOKER, "readToken(\\n): `" +sbuf+ "'");
					return sbuf.toString();
				}
				if (Character.isWhitespace(c))
				{
					this.eol = false;
					Trace.t(Trace.TOKER, "readToken(ws): `" +sbuf+ "'");
					if (sbuf.length() > 0)
						return sbuf.toString();
					// eat useless leading whitespace
				}
				this.eol = false;
				sbuf.append(c);
			}
		}
		catch (IOException ioe)
		{
			throw new Error("Gah.", ioe);
		}
	}

	void lineDone()
	{
		try
		{
			if (this.eol)
			{
				this.eol = false;
				return;
			}

			while (true)
			{
				int ci = this.input.read();
				if (ci == -1)
					return;
				char c = (char)ci;
				if (c == '\n')
				{
					this.eol = false;
					return;
				}
				Trace.t(Trace.TOKER, "lineDone() skipping: `" +c+ "'");
				this.eol = false;
			}
		}
		catch (IOException ioe)
		{
			throw new Error("Gah.", ioe);
		}
		finally
		{
			// Finished with the line
			this.eol = false;
		}
	}

	void flush()
	{
		try
		{
			while (true)
			{
				int ci = this.input.read();
				if (ci == -1)
					return;
				char c = (char)ci;
				System.err.print(c);
			}
		}
		catch (IOException ioe)
		{
			System.err.println("ioe during flush: " +ioe);
		}
	}

	
	boolean testEndOfLine()
	{
		try
		{
			int ci = this.input.read();
			Trace.t(Trace.TOKER, "testEOL() checking: `" +ci+ "'");
			if (ci == -1)
			{
				this.eol = true;
				return true;
			}
			char c = (char)ci;
			if (c == '\n')
			{
				this.eol = true;
				return true;
			}
			this.input.unread(ci);
			this.eol = false;
			return false;
		}
		catch (IOException ioe)
		{
			throw new Error("Gah.", ioe);
		}
	}


	boolean endOfLine()
	{
		return this.eol;
	}
	
}
