/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.util.Iterator;
import java.io.PrintStream;

public final class WorldState
{
	private final Robot me;

	// Integer id -> OtherRobot map
	private IntKeyedHashtable robots = new IntKeyedHashtable(50);

	// Integer id -> Package map
	private IntKeyedHashtable packages = new IntKeyedHashtable(50);

	public static class OtherRobot
	implements IntKeyedHashtable.IntKey
	{
		final int id;
		int x;
		int y;
		boolean stale; // For GC'ing stale robots

		public OtherRobot(int id)
		{
			this.id = id;
			this.x = this.y = -1;
		}

		public void placed(int x, int y)
		{
			this.x = x;
			this.y = y;
		}

		public void picked(int pkg)
		{
		}

		public void dropped(int pkg)
		{
		}

		public void moved(Direction d)
		{
			if (knownLocation())
			{
				this.x += d.dx();
				this.y += d.dy();
			}
		}

		public boolean knownLocation()
		{
			return (this.x != -1);
		}

		public int keyInt()
		{
			return this.id;
		}

		public String toString()
		{
			return "#" +this.id+ " (" +x+ "," +y+ ")";
		}

		public int hashCode()
		{
			return this.id;
		}

		public boolean equals(Object other)
		{
			if (other instanceof OtherRobot)
			{
				return ((OtherRobot)other).id == this.id;
			}
			return false;
		}
	}

	public WorldState(Robot me)
	{
		this.me = me;
	}

	public Package getPackage(int id, Package defaultPkg)
	{
		if ((defaultPkg != null)
		    && (id != defaultPkg.id()))
			throw new Error("Lookup id inconsistent with default package");

		Package p = (Package)this.packages.get(id);
		if (p == null)
		{
			if (defaultPkg == null)
				defaultPkg = new Package(id);
			this.packages.add(defaultPkg);
			p = defaultPkg;
		}
		return p;
	}

	public void removePackage(Package p)
	{
		//this.packages.remove(p);
	}

	public OtherRobot lookupRobot(int id)
	{
		OtherRobot r = (OtherRobot)robots.get(id);
		if (r == null)
		{
			r = new OtherRobot(id);
			robots.add(r);
		}
		r.stale = false;
		return (OtherRobot)r;
	}

	public void updatePackages(final Package[] currPkgs, int x, int y)
	{
		for (int i = 0; i < currPkgs.length; i++)
		{
			Package p = currPkgs[i];
			Package existing = getPackage(p.id(), /*default:*/p);
			if (p != existing)
				existing.update(p);
			existing.setLocation(x,y);
		}
	}


	public void robotPlaced(int id, int x, int y)
	{
		if (id == me.id())
			me.placed(x,y);
		else
		{
			OtherRobot r = lookupRobot(id);
			r.placed(x,y);
		}
	}

	public void robotPicked(int robotId, int pkgId)
	{
		// Add to known packages, if necessary
		Package p = getPackage(pkgId, null);

		// Update package location (picked up)
		p.setLocation(-1, -1);

		if (robotId == me.id())
			me.picked(p);
		else
		{
			OtherRobot r = lookupRobot(robotId);
			r.picked(pkgId);
		}
	}

	public void robotDropped(int robotId, int pkgId)
	{
		// Add to known packages, if necessary
		Package p = getPackage(pkgId, null);
		int x, y;

		if (robotId == me.id())
		{
			me.dropped(p);
			x = me.x();
			y = me.y();
		}
		else
		{
			OtherRobot r = lookupRobot(robotId);
			r.dropped(pkgId);
			x = r.x;
			y = r.y;
		}

		p.setLocation(x, y);
		if (p.delivered())
			removePackage(p);
	}

	public void robotMoved(int robotId, Direction d)
	{
		if (robotId == me.id())
			me.moved(d);
		else
		{
			OtherRobot r = lookupRobot(robotId);
			r.moved(d);
		}
	}

	public void markOthers()
	{
		Iterator i = robots.iterator();
		while(i.hasNext())
		{
			OtherRobot r = (OtherRobot)(i.next());
			r.stale = true;
		}
	}

	public void printOthers(PrintStream ps)
	{
		boolean p = false;
		Iterator i = robots.iterator();
		while(i.hasNext())
		{
			OtherRobot r = (OtherRobot)(i.next());
			if (!r.stale)
			{
				p = true;
				ps.print(r.toString() + " ");
			}
		}
		if (p)
			ps.println();
	}

	public Package pickRandomKnownPackage()
	{
		Iterator i = packages.iterator();
		while (i.hasNext())
		{
			Package p = (Package)(i.next());
			if ((!p.delivered())
			    && (p.x() != -1)
			    && (p.y() != -1))
				return p;
		}
		return null;
	}

	public OtherRobot pickRandomKnownRobot()
	{
		Iterator i = robots.iterator();
		while (i.hasNext())
		{
			OtherRobot r = (OtherRobot)(i.next());
			if ((!r.stale)
			    && (r.x != -1)
			    && (r.y != -1))
				return r;
		}
		return null;
	}

	
	public void sweepOthers()
	{
//		Iterator i = robots.iterator();
//		while(i.hasNext())
//		{
//			OtherRobot r = (OtherRobot)(i.next());
//			r.dead = true;
//		}
	}
	
}

